/********* Sequence input routines for CLUSTAL W *******************/
/* DES was here.  FEB. 1994 */
/* Now reads PILEUP/MSF and CLUSTAL alignment files */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <math.h>
#include "clustalw.h"	

#define MIN(a,b) ((a)<(b)?(a):(b))



/*
*	Prototypes
*/

static DCHAR * get_seq(char *,sint *,char *);
static DCHAR * get_clustal_seq(char *,sint *,char *,sint);
static DCHAR * get_msf_seq(char *,sint *,char *,sint);
static void check_infile(sint *nseqs, Boolean append);
static void p_encode(DCHAR *, ECHAR *, sint);
static ECHAR res_index(DCHAR *,DCHAR);
static sint count_clustal_seqs(Boolean append);
static sint count_gde_seqs(Boolean append);
static sint count_pir_seqs(Boolean append);
static sint count_swiss_seqs(Boolean append);
static sint count_pearson_seqs(Boolean append);
static sint count_msf_seqs(Boolean append);
static sint count_rsf_seqs(Boolean append);
static void get_swiss_feature(char *line);
static void get_rsf_feature(char *line);
static void get_swiss_mask(char *line);
static void get_clustal_ss(sint length);
static void get_embl_ss(sint length);
static void get_rsf_ss(sint length);
static void get_gde_ss(sint length);
static Boolean cl_blankline(char *line);
static void init_alphabet(void);
static sint update_alphabet(char *line,sint s);
static void update_seq(char *line,DCHAR *seq,sint *len);
static void update_ss(char *line,char *ss,sint *len);
static Boolean isgap(char c);

/*
 *	Global variables
 */
extern sint nalphabet;
extern sint max_names;
extern FILE *fin;
extern Boolean usemenu;
extern Boolean interactive;
extern char seqname[];
extern sint nseqs;
extern sint *seqlen_array;
extern sint *output_index;
extern char **names,**titles;
extern ECHAR **seq_array;
extern Boolean profile1_empty, profile2_empty;
extern sint max_aln_length;
extern char *gap_penalty_mask;
extern sint struct_penalties;
extern char *ss_name;
extern sint profile_no;
extern sint debug;
extern sint max_aa,gap_pos1,gap_pos2;
extern sint GAPCOL,LENCOL;

DCHAR *alphabet;  /* DES */
static sint seqFormat;
static char *formatNames[] = {"unknown","EMBL/Swiss-Prot","PIR",
                              "Pearson","GDE","Clustal","Pileup/MSF","RSF","USER"};
static sint line_no;



static Boolean cl_blankline(char *line)
{
	int i;

	if (line[0] == '!') return TRUE;
	
	for(i=0;line[i]!='\n' && line[i]!=EOS;i++) {
		if( isdigit(line[i]) ||
		    isspace(line[i]) ||
		    (line[i] == '#') ||
		    (line[i] == '*') ||
		    (line[i] == ':') ||
                    (line[i] == '.')) 
			;
		else
			return FALSE;
	}
	return TRUE;
}

static DCHAR * get_msf_seq(char *sname,sint *len,char *tit,sint seqno)
/* read the seqno_th. sequence from a PILEUP multiple alignment file */
{
        static char line[MAXLINE+1];
        DCHAR *seq = NULL;
        sint i,j,k;
        DCHAR c;

        fseek(fin,0,0);                 /* start at the beginning */

        *len=0;                         /* initialise length to zero */
        for(i=0;;i++) {
                if(fgets(line,MAXLINE+1,fin)==NULL) return NULL; /* read the title*/
                if(linetype(line,"//") ) break;             /* lines...ignore*/
        }

        while (fgets(line,MAXLINE+1,fin) != NULL) {
                if(!blankline(line)) {

                        for(i=1;i<seqno;i++) {
				fgets(line,MAXLINE+1,fin);
			}
                        for(j=0;j<=strlen(line);j++) if(line[j] != ' ') break;
                        for(k=j;k<=strlen(line);k++) if(line[k] == ' ') break;
                        strncpy(sname,line+j,MIN(MAXNAMES,k-j));
                        sname[MIN(MAXNAMES,k-j)]=EOS;
                        rtrim(sname);
                        blank_to_(sname);

                        if(seq==NULL)
                                seq=(DCHAR *)ckalloc((MAXLINE+2)*sizeof(DCHAR));
                        else
                                seq=(DCHAR *)ckrealloc(seq,((*len)+MAXLINE+2)*sizeof(DCHAR));
			update_seq(&line[k],seq,len);

                        for(i=0;;i++) {
                                if(fgets(line,MAXLINE+1,fin)==NULL) return seq;
                                if(blankline(line)) break;
                        }
                }
        }
	seq[(*len)+1].c[0]=EOS;
        return seq;
}

static DCHAR * get_clustal_seq(char *sname,sint *len,char *tit,sint seqno)
/* read the seqno_th. sequence from a clustal multiple alignment file */
{
	static char line[MAXLINE+1];
	static char tseq[MAXLINE+1];
	DCHAR *seq = NULL;
	sint i,j;
	DCHAR c;

	fseek(fin,0,0); 		/* start at the beginning */

	*len=0;				/* initialise length to zero */
	fgets(line,MAXLINE+1,fin);	/* read the title line...ignore it */

	while (fgets(line,MAXLINE+1,fin) != NULL) {
		if(!cl_blankline(line)) {

			for(i=1;i<seqno;i++) fgets(line,MAXLINE+1,fin);
			for(j=0;j<=strlen(line);j++) if(line[j] != ' ') break;

			sscanf(line,"%s%s",sname,tseq);
			for(j=0;j<MAXNAMES;j++) if(sname[j] == ' ') break;
			sname[j]=EOS;
			rtrim(sname);
                       	blank_to_(sname);

			if(seq==NULL)
				seq=(DCHAR *)ckalloc((MAXLINE+2)*sizeof(DCHAR));
			else
				seq=(DCHAR *)ckrealloc(seq,((*len)+MAXLINE+2)*sizeof(DCHAR));
			update_seq(tseq,seq,len);
			for(i=0;;i++) {
				if(fgets(line,MAXLINE+1,fin)==NULL) return seq;
				if(cl_blankline(line)) break;
			}
		}
	}
	seq[(*len)+1].c[0]=EOS;

	return seq;
}

static void get_clustal_ss(sint length)
/* read the structure data from a clustal multiple alignment file */
{
	static char title[MAXLINE+1];
	static char line[MAXLINE+1];
	static char lin2[MAXLINE+1];
	static char tseq[MAXLINE+1];
	static char sname[MAXNAMES+1];
	sint i,j,len,ix,struct_index=0;
	char c;

	
	fseek(fin,0,0); 		/* start at the beginning */

	len=0;				/* initialise length to zero */
	if (fgets(line,MAXLINE+1,fin) == NULL) return;	/* read the title line...ignore it */

	if (fgets(line,MAXLINE+1,fin) == NULL) return;  /* read the next line... */
/* skip any blank lines */
	for (;;) {
		if(fgets(line,MAXLINE+1,fin)==NULL) return;
		if(!blankline(line)) break;
	}

/* look for structure table lines */
	ix = -1;
	for(;;) {
		if(line[0] != '!') break;
		if(strncmp(line,"!GM",3) == 0) {
			ix++;
			sscanf(line+4,"%s%s",sname,tseq);
			for(j=0;j<MAXNAMES;j++) if(sname[j] == ' ') break;
			sname[j]=EOS;
			rtrim(sname);
    		blank_to_(sname);
    		if (interactive) {
				strcpy(title,"Found gap penalty mask in alignment file: ");
				strcat(title,sname);
				(*lin2)=prompt_for_yes_no(title,"Use it to set local gap penalties ");
			}
			else (*lin2) = 'y';
			if ((*lin2 != 'n') && (*lin2 != 'N'))  {               	
				struct_penalties = GMASK;
				struct_index = ix;
				for (i=0;i<length;i++)
					gap_penalty_mask[i] = '1';
				strcpy(ss_name,sname);
				update_ss(tseq,gap_penalty_mask,&len);
			}
		}
		if (struct_penalties != NONE) break;
		if(fgets(line,MAXLINE+1,fin)==NULL) return;
	}
			
	if (struct_penalties == NONE) return;
	
/* skip any more comment lines */
	while (line[0] == '!') {
		if(fgets(line,MAXLINE+1,fin)==NULL) return;
	}

/* skip the sequence lines and any comments after the alignment */
	for (;;) {
		if(isspace(line[0])) break;
		if(fgets(line,MAXLINE+1,fin)==NULL) return;
	}
			

/* read the rest of the alignment */
	
	for (;;) {
/* skip any blank lines */
			for (;;) {
				if(!isspace(line[0])) break;
				if(fgets(line,MAXLINE+1,fin)==NULL) return;
			}
/* get structure table line */
			for(ix=0;ix<struct_index;ix++) {
				if (line[0] != '!') {
					error("bad gap penalty mask format");
				   	struct_penalties = NONE;
					return;
				}
				if(fgets(line,MAXLINE+1,fin)==NULL) return;
			}
			if (struct_penalties == GMASK) {
				if (strncmp(line,"!GM",3) != 0) {
					error("bad gap penalty mask format");
					struct_penalties = NONE;
					return;
				}
				sscanf(line+4,"%s%s",sname,tseq);
				update_ss(tseq,gap_penalty_mask,&len);
			}

/* skip any more comment lines */
		while (line[0] == '!') {
			if(fgets(line,MAXLINE+1,fin)==NULL) return;
		}

/* skip the sequence lines */
		for (;;) {
			if(isspace(line[0])) break;
			if(fgets(line,MAXLINE+1,fin)==NULL) return;
		}
	}
}


static void get_gde_ss(sint length)
{
	static char title[MAXLINE+1];
	static char line[MAXLINE+1];
	static char lin2[MAXLINE+1];
	static char sname[MAXNAMES+1];
	sint i, len, offset = 0;
        unsigned char c;

	for (;;) {
		line[0] = '\0';
/* search for the next comment line */
		while(*line != '"')
			if (fgets(line,MAXLINE+1,fin) == NULL) return;

/* is it a gap penalty mask entry? */
		if (strncmp(&line[1],"GM_",3) == 0) {
			for (i=1;i<=MAXNAMES-3;i++) {
				if (line[i+3] == '(' || line[i+3] == '\n')
						break;
				sname[i-1] = line[i+3];
			}
			i--;
			sname[i]=EOS;
			if (sname[i-1] == '(') sscanf(&line[i+3],"%d",&offset);
			else offset = 0;
			for(i--;i > 0;i--) 
				if(isspace(sname[i])) {
					sname[i]=EOS;	
				}
				else break;		
			blank_to_(sname);

			if (interactive) {
				strcpy(title,"Found gap penalty mask in alignment file: ");
				strcat(title,sname);
				(*lin2)=prompt_for_yes_no(title,"Use it to set local gap penalties ");
			}
			else (*lin2) = 'y';
			if ((*lin2 != 'n') && (*lin2 != 'N'))  {               	
				struct_penalties = GMASK;
				for (i=0;i<length;i++)
					gap_penalty_mask[i] = '1';
				len = 0;
				while(fgets(line,MAXLINE+1,fin)) {
					if(*line == '%' || *line == '#' || *line == '"') break;
					for(i=offset;i < length;i++) {
						c=line[i];
						if(c == '\n' || c == EOS) 
							break;			/* EOL */
						gap_penalty_mask[len++]=c;
					}
					if (len > length) break;
				}
				strcpy(ss_name,sname);
			}
		}
		if (struct_penalties != NONE) break;			
	}			
			
}


static DCHAR * get_seq(char *sname,sint *len,char *tit)
{
	static char line[MAXLINE+1];
	DCHAR *seq = NULL;
	sint i,j;
    DCHAR c;
	Boolean got_seq=FALSE;

	switch(seqFormat) {
/************************************/
                case EMBLSWISS:
                        while( !linetype(line,"ID") )
                                if (fgets(line,MAXLINE+1,fin) == NULL) return NULL;

                        for(i=5;i<=strlen(line);i++)  /* DES */
                                if(line[i] != ' ') break;
                        strncpy(sname,line+i,MAXNAMES); /* remember entryname */
                        for(i=0;i<=strlen(sname);i++)
                                if(sname[i] == ' ') {
                                        sname[i]=EOS;
                                        break;
                                }

                        sname[MAXNAMES]=EOS;
                        rtrim(sname);
                        blank_to_(sname);


                        while( !linetype(line,"SQ") )
                                fgets(line,MAXLINE+1,fin);

                        *len=0;
                        while(fgets(line,MAXLINE+1,fin)) {
                                if( strlen(line) > 2 && line[strlen(line)-2]=='.' && line[strlen(line)-3]=='.' )
                                        continue;
                                if(seq==NULL)
                                        seq=(DCHAR *)ckalloc((MAXLINE+2)*sizeof(DCHAR));
                                else
                                        seq=(DCHAR *)ckrealloc(seq,((*len)+MAXLINE+2)*sizeof(DCHAR));
				if(linetype(line,"ID"))
					break;			/* EOL */
				update_seq(line,seq,len);
                        if(line[i] == '/') break;
                        }
                break;

/************************************/
                case PIR:
                        while(*line != '>')
                                fgets(line,MAXLINE+1,fin);
                        for(i=4;i<=strlen(line);i++)  /* DES */
                                if(line[i] != ' ') break;
                        strncpy(sname,line+i,MAXNAMES); /* remember entryname */
                        sname[MAXNAMES]=EOS;
                        rtrim(sname);
                        blank_to_(sname);

                        fgets(line,MAXLINE+1,fin);
                        strncpy(tit,line,MAXTITLES);
                        tit[MAXTITLES]=EOS;
                        i=strlen(tit);
                        if(tit[i-1]=='\n') tit[i-1]=EOS;

                        *len=0;
                        while(fgets(line,MAXLINE+1,fin)) {
                                if(seq==NULL)
                                        seq=(DCHAR *)ckalloc((MAXLINE+2)*sizeof(DCHAR));
                                else
                                        seq=(DCHAR *)ckrealloc(seq,((*len)+MAXLINE+2)*sizeof(DCHAR));
				update_seq(line,seq,len);
                        	if(line[0] == '*') break;
                        }
                break;
/***********************************************/
                case PEARSON:
                        while(*line != '>')
                                fgets(line,MAXLINE+1,fin);

                        for(i=1;i<=strlen(line);i++)  /* DES */
                                if(line[i] != ' ') break;
                        strncpy(sname,line+i,MAXNAMES); /* remember entryname */
                        for(i=1;i<=strlen(line);i++)  /* DES */
                                if(sname[i] == ' ') break;
                        sname[i]=EOS;
                        rtrim(sname);
                        blank_to_(sname);

                        *tit=EOS;

                        *len=0;
                        while(fgets(line,MAXLINE+1,fin)) {
				if(*line == '>') break;
                                if(seq==NULL)
                                        seq=(DCHAR *)ckalloc((MAXLINE+2)*sizeof(DCHAR));
                                else
                                        seq=(DCHAR *)ckrealloc(seq,((*len)+MAXLINE+2)*sizeof(DCHAR));
				update_seq(line,seq,len);
                        	if(line[0] == '>') break;
                        }
                break;

/**********************************************/
		case GDE:
				while(*line != '%')
					fgets(line,MAXLINE+1,fin);
			
			for (i=1;i<=MAXNAMES;i++) {
				if (line[i] == '\n')
                                    break;
				sname[i-1] = line[i];
			}
			i--;
			sname[i]=EOS;
			for(i--;i > 0;i--) 
				if(isspace(sname[i])) {
					sname[i]=EOS;	
				}
				else break;		
                        blank_to_(sname);

			*tit=EOS;
			
			*len=0;
			while(fgets(line,MAXLINE+1,fin)) {
			if(*line == '%' || *line == '#' || *line == '"') break;
				if(seq==NULL)
					seq=(DCHAR *)ckalloc((MAXLINE+2)*sizeof(DCHAR));
				else
					seq=(DCHAR *)ckrealloc(seq,((*len)+MAXLINE+2)*sizeof(DCHAR));
				update_seq(line,seq,len);
			}
		break;
/***********************************************/
                case RSF:
                        while(*line != '{')
                                if (fgets(line,MAXLINE+1,fin) == NULL) return NULL;

                        while( !keyword(line,"name") )
                                if (fgets(line,MAXLINE+1,fin) == NULL) return NULL;

                        for(i=5;i<=strlen(line);i++)  /* DES */
                                if(line[i] != ' ') break;
                        strncpy(sname,line+i,MAXNAMES); /* remember entryname */
                        for(i=0;i<=strlen(sname);i++)
                                if(sname[i] == ' ') {
                                        sname[i]=EOS;
                                        break;
                                }

                        sname[MAXNAMES]=EOS;
                        rtrim(sname);
                        blank_to_(sname);


                        while( !keyword(line,"sequence") )
                                if (fgets(line,MAXLINE+1,fin) == NULL) return NULL;

                        *len=0;
                        while(fgets(line,MAXLINE+1,fin)) {
                                if(seq==NULL)
                                        seq=(DCHAR *)ckalloc((MAXLINE+2)*sizeof(DCHAR));
                                else
                                        seq=(DCHAR *)ckrealloc(seq,((*len)+MAXLINE+2)*sizeof(DCHAR));
				if(line[0] == '}' || line[0] == EOS) break;
				update_seq(line,seq,len);
                        }
                break;
/***********************************************/

	}
	
	seq[*len+1].c[0]=EOS;

	return seq;
}

static void update_seq(char *line,DCHAR *seq,sint *len)
{
	sint i,j;
	DCHAR c;
	Boolean eol=FALSE;

	for(i=0;i<=MAXLINE;) {
		for(j=0;j<nalphabet;)
		{
			if(line[i]=='\n' || line[i] == EOS) {
				eol=TRUE;
				break; /* EOL */
			}
			if (isspace(line[i])) {
				i++;
				continue;
			}
			c.c[j++]=line[i++];
		}
		if(eol) break;

		if(isalpha(c.c[0]))
			seq[++(*len)]=cpchar(c);
		else if(!isdigit(c.c[0]))
			seq[++(*len)]=gapchar();

	}
}

static void update_ss(char *line,char *ss,sint *len)
{
	sint i,j;
	DCHAR c;
	Boolean eol=FALSE;

	for(i=0;i<=MAXLINE;) {
		for(j=0;j<nalphabet;)
		{
			if(line[i]=='\n' || line[i] == EOS) {
				eol=TRUE;
				break; /* EOL */
			}
			if (isspace(line[i])) {
				i++;
				continue;
			}
			c.c[j++]=line[i++];
		}
		if(eol) break;

		if(isdigit(c.c[0]))
			ss[++(*len)]=c.c[0];
		else if(!isspace(c.c[0]))
			ss[++(*len)]='-';

	}
}

sint readseqs(sint first_seq,Boolean append) /*first_seq is the #no. of the first seq. to read */
{
	char line[FILENAMELEN+1];
	static char sname1[MAXNAMES+1],title[MAXTITLES+1];
	static DCHAR *seq1;
	sint i,j;
	sint no_seqs;
	static sint l1;
	
	if(usemenu)
		getstr("Enter the name of the sequence file",line);
	else
		strcpy(line,seqname);
	if(*line == EOS) return -1;
	
	if((fin=fopen(line,"r"))==NULL) {
		error("Could not open sequence file %s",line);
		return -1;      /* DES -1 => file not found */
	}
	strcpy(seqname,line);
	no_seqs=0;
	check_infile(&no_seqs,append);

/* DES DEBUG 
	fprintf(stdout,"\n\n File name = %s\n\n",seqname);
*/
	if(no_seqs <= 0)
		return no_seqs;      /* return the number of seqs. (zero here)*/

	if(max_aa <= 0)
		return max_aa;

	info("Sequence format is %s",formatNames[seqFormat]);
		
	gap_pos1=max_aa;
	gap_pos2=gap_pos1+1;
	GAPCOL=gap_pos2+1;
	LENCOL=GAPCOL+1;

	max_aln_length = 0;

/* if this is a multiple alignment, or profile 1 - free any memory used
by previous alignments, then allocate memory for the new alignment */
	if(first_seq == 1) {
		max_names = 0;
		free_aln(nseqs);
		alloc_aln(no_seqs);
	}
/* otherwise, this is a profile 2, and we need to reallocate the arrays,
leaving the data for profile 1 intact */
	else realloc_aln(first_seq,no_seqs);

        for(i=1;i<first_seq;i++)
	{
                if(seqlen_array[i]>max_aln_length)
                        max_aln_length=seqlen_array[i];
		if(strlen(names[i])>max_names)
			max_names=strlen(names[i]);
	}

	for(i=first_seq;i<=first_seq+no_seqs-1;i++) {    /* get the seqs now*/
		output_index[i] = i;	/* default output order */
                if(seqFormat == CLUSTAL)
                        seq1=get_clustal_seq(sname1,&l1,title,i-first_seq+1);
                else if(seqFormat == MSF)
                            seq1=get_msf_seq(sname1,&l1,title,i-first_seq+1);
                else
                        seq1=get_seq(sname1,&l1,title);


		if(seq1==NULL) break;
/* JULIE */
/*  Set max length of dynamically allocated arrays in prfalign.c */
		if (l1 > max_aln_length) max_aln_length = l1;
		seqlen_array[i]=l1;                   /* store the length */
		strcpy(names[i],sname1);              /*    "   "  name   */
		strcpy(titles[i],title);              /*    "   "  title  */

		alloc_seq(i,l1);

			p_encode(seq1,seq_array[i],l1);
		if(seq1!=NULL) seq1=ckfree(seq1);
	}

	max_aln_length *= 2;
/*
   JULIE
   check sequence names are all different - otherwise phylip tree is 
   confused.
*/
	for(i=1;i<=first_seq+no_seqs-1;i++) {
		for(j=i+1;j<=first_seq+no_seqs-1;j++) {
			if (strncmp(names[i],names[j],MAXNAMES) == 0) {
				error("Multiple sequences found with same name, %s (first %d chars are significant)", names[i],MAXNAMES);
				return -1;
			}
		}
	}
	for(i=first_seq;i<=first_seq+no_seqs-1;i++)
	{
		if(seqlen_array[i]>max_aln_length)
			max_aln_length=seqlen_array[i];
	}
	
/* look for a feature table / gap penalty mask (only if this is a profile) */
	if (profile_no > 0) {
		rewind(fin);
		struct_penalties = NONE;
    		gap_penalty_mask = (char *)ckalloc((max_aln_length+1) * sizeof (char));
		ss_name = (char *)ckalloc((MAXNAMES+1) * sizeof (char));

		if (seqFormat == CLUSTAL) {
			get_clustal_ss(max_aln_length);
		}
		else if (seqFormat == GDE) {
			get_gde_ss(max_aln_length);
		}
	}

	for(i=first_seq;i<=first_seq+no_seqs-1;i++)
	{
		if(strlen(names[i])>max_names)
			max_names=strlen(names[i]);
	}

	if(max_names<10) max_names=10;

	fclose(fin);
			
	return no_seqs;    /* return the number of seqs. read in this call */
}


static void check_infile(sint *nseqs,Boolean append)
{
	char line[MAXLINE+1];
	sint i;	

	*nseqs=0;
	line_no=0;
	while (fgets(line,MAXLINE+1,fin) != NULL) {
		line_no++;
		if(!blankline(line)) 
			break;
	}

	for(i=strlen(line)-1;i>=0;i--)
		if(isgraph(line[i])) break;
	line[i+1]=EOS;
        
	for(i=0;i<=6;i++) line[i] = toupper(line[i]);

       if( linetype(line,"ID") ) {                                     /* EMBL/Swiss-Prot format ? */
                seqFormat=EMBLSWISS;
                (*nseqs)++;
        }
        else if( linetype(line,"CLUSTAL") ) {
                seqFormat=CLUSTAL;
        }
        else if( linetype(line,"PILEUP") ) {
                seqFormat = MSF;
        }
        else if( linetype(line,"!!AA_MULTIPLE_ALIGNMENT") ) {
                seqFormat = MSF;
        }
        else if( linetype(line,"!!NA_MULTIPLE_ALIGNMENT") ) {
                seqFormat = MSF;
        }
        else if( strstr(line,"MSF") && line[strlen(line)-1]=='.' &&
                                 line[strlen(line)-2]=='.' ) {
                seqFormat = MSF;
        }
        else if( linetype(line,"!!RICH_SEQUENCE") ) {
                seqFormat = RSF;
        }
        else if(*line == '>') {                                         /* no */
                seqFormat=(line[3] == ';')?PIR:PEARSON; /* distinguish PIR and Pearson */
                (*nseqs)++;
        }
        else if((*line == '"') || (*line == '%') || (*line == '#')) {
                seqFormat=GDE; /* GDE format */
                if (*line == '%') {
                        (*nseqs)++;
                }
                else if (*line == '#') {
                        (*nseqs)++;
                }
        }
        else {
                seqFormat=UNKNOWN;
                return;
        }


                switch(seqFormat) {
                        case EMBLSWISS:
                                *nseqs = count_swiss_seqs(append);
                                break;
                        case PIR:
                                *nseqs = count_pir_seqs(append);
                                fseek(fin,0,0);
                                break;
                        case PEARSON:
                                *nseqs = count_pearson_seqs(append);
                                fseek(fin,0,0);
                                break;
                        case GDE:
				*nseqs = count_gde_seqs(append);
                                break;
                        case CLUSTAL:
                                *nseqs = count_clustal_seqs(append);
/* DES */                       /* fprintf(stdout,"\nnseqs = %d\n",(pint)*nseqs); */
                                break;
                        case MSF:
                                *nseqs = count_msf_seqs(append);
                                break;
                        case RSF:
                                fseek(fin,0,0);
                                *nseqs = count_rsf_seqs(append);
                                break;
                        case USER:
                        default:
                                break;
                }

	if (*nseqs <= 0) return;
	fseek(fin,0,0);
}


static void init_alphabet(void)
{
	int n;

	n=(int)pow((double)27,(double)nalphabet);
        max_aa=0;
        if(alphabet!=NULL) ckfree(alphabet);
        alphabet=(DCHAR *)ckalloc(n*sizeof(DCHAR));
}

static sint update_alphabet(char *line,sint s)
{
	sint i,j,k,ix;
        Boolean found;
	DCHAR c;
	Boolean gap;

        for(i=s;i<=MAXLINE;) {
        	if(line[i] == '\n' || line[i] == EOS)
                	break;                  /* EOL */
/* get the first character and decide if it's a letter or a gap */
        	for(j=0;;i++)
        	{
        		if(line[i] == '\n' || line[i] == EOS)
                		break;                  /* EOL */
                	if (isalpha(line[i])) {
				gap=FALSE;
				break;
			}
                	else if (isgap(line[i])) {
				gap=TRUE;
				break;
			}
			else if(!isspace)
				fprintf(stdout,"got %d\n",(int)line[i]);
        	}
        	if(line[i] == '\n' || line[i] == EOS)
                	break;                  /* EOL */
                c.c[j++]=line[i++];
/* now get the rest of the character */
        	for(;j<nalphabet;)
        	{
        		if(line[i] == '\n' || line[i] == EOS)
                		break;                  /* EOL */
                	if (isspace(line[i])) {
                        	i++;
                        	continue;
			}
			else if(isalpha(line[i]) && gap==FALSE) {
                		c.c[j++]=line[i++];
			}
			else if(isgap(line[i]) && gap==TRUE) {
                		i++;
				j++;
			}
			else {
				error("Wrong format in line number: %d\n"
					"Sequences are not compatible with alphabet size\n",line_no);
				return (sint)-1;
			}
        	}
		if(gap==FALSE) {
        		found=FALSE;
        		for(k=0;k<max_aa;k++) {
                		if (cmpchar(c,alphabet[k])) {
                        		found=TRUE;
                        		break;
                		}
        		}
        		if(found==FALSE) {
                		alphabet[max_aa]=cpchar(c);
                		max_aa++;
        		}
		}
	}
	return (sint)1;
}

static Boolean isgap(char c)
{
	Boolean ret;
	if(c=='-'||c=='.'||c=='~') ret=TRUE;
	else ret=FALSE;

	return ret;
}

static sint count_swiss_seqs(Boolean append)
/* count the number of sequences in a clustal alignment file */
{
        sint  nseqs;
        static char line[MAXLINE+1];
        static char sname[MAXLINE+1];
        sint i,j;

        nseqs=1;
        while(fgets(line,MAXLINE+1,fin)) {
		line_no++;
                if( linetype(line,"ID") )
                        nseqs++;
        }

        if (nseqs==0) return nseqs;

/* setup the alphabet now */
	if (!append) init_alphabet();


        fseek(fin,0,0);                 /* start at the beginning */
	line_no=0;

        for(;;) {
                while( !linetype(line,"SQ") )
                        if(fgets(line,MAXLINE+1,fin)==NULL) return nseqs;

                while(fgets(line,MAXLINE+1,fin)) {
			line_no++;
                        if( strlen(line) > 2 && line[strlen(line)-2]=='.' && line[strlen(line)-3]=='.' )
                                       continue;
                        if(linetype(line,"ID")) break;
			if(blankline(line)) continue;		/* Look for next non- */
			if(update_alphabet(line,0)<=0) return (sint)-1;
                }
        }

}
static sint count_gde_seqs(Boolean append)
/* count the number of sequences in a clustal alignment file */
{
        sint  nseqs;
        static char line[MAXLINE+1];
        static char sname[MAXLINE+1];
        sint i,j;

        nseqs=1;
        while(fgets(line,MAXLINE+1,fin)) {
			line_no++;
                if(*line == '%') nseqs++;
        }

        if (nseqs==0) return nseqs;

/* setup the alphabet now */
	if(!append) init_alphabet();


        fseek(fin,0,0);                 /* start at the beginning */
	line_no=0;

        for(;;) {
                while(*line != '%')
                        if(fgets(line,MAXLINE+1,fin)==NULL) return nseqs;

                while(fgets(line,MAXLINE+1,fin)) {
			line_no++;
                        if(*line == '%' || *line == '#' || *line == '"') break;
			if(update_alphabet(line,0)<=0) return (sint)-1;
                }
        }

}



static sint count_pearson_seqs(Boolean append)
/* count the number of sequences in a pir alignment file */
{
	char line[MAXLINE+1];
	sint  nseqs, i, j;
	Boolean seq_ok;

	nseqs=1;
	while (fgets(line,MAXLINE+1,fin) != NULL) { /* Look for end of first seq */
			line_no++;
		if(*line == '>') nseqs++;
	}

        if (nseqs==0) return nseqs;

/* setup the alphabet now */
	if (!append) init_alphabet();


        fseek(fin,0,0);                 /* start at the beginning */
	line_no=0;

        for(;;) {
		while(*line != '>')
                	if(fgets(line,MAXLINE+1,fin)==NULL) return nseqs;

                while(fgets(line,MAXLINE+1,fin)) {
			line_no++;
                        if(*line == '>') break;
			if(update_alphabet(line,0)<=0) return (sint)-1;
                }
        }
}


static sint count_pir_seqs(Boolean append)
/* count the number of sequences in a pir alignment file */
{
	char line[MAXLINE+1],c;
	sint  nseqs, i, j;
	Boolean seq_ok;

	seq_ok = FALSE;
	while (fgets(line,MAXLINE+1,fin) != NULL) { /* Look for end of first seq */
			line_no++;
		if(*line == '>') break;
		for(i=0;seq_ok == FALSE;i++) {
			c=line[i];
			if(c == '*') {
				seq_ok = TRUE;	/* ok - end of sequence found */
				break;
			}			/* EOL */
			if(c == '\n' || c == EOS)
				break;			/* EOL */
		}
		if (seq_ok == TRUE)
			break;
	}
	if (seq_ok == FALSE) {
		error("PIR format sequence end marker '*'\nmissing for one or more sequences.");
		return (sint)0;	/* funny format*/
	}
	
	
	nseqs = 1;
	
	while (fgets(line,MAXLINE+1,fin) != NULL) {
			line_no++;
		if(*line == '>') {		/* Look for start of next seq */
			seq_ok = FALSE;
			while (fgets(line,MAXLINE+1,fin) != NULL) { /* Look for end of seq */
			line_no++;
				if(*line == '>') {
					error("PIR format sequence end marker '*' missing for one or more sequences.");
					return (sint)0;	/* funny format*/
				}
				for(i=0;seq_ok == FALSE;i++) {
					c=line[i];
					if(c == '*') {
						seq_ok = TRUE;	/* ok - sequence found */
						break;
					}			/* EOL */
					if(c == '\n' || c == EOS)
						break;			/* EOL */
				}
				if (seq_ok == TRUE) {
					nseqs++;
					break;
				}
			}
		}
	}

        if (nseqs==0) return nseqs;

/* setup the alphabet now */
	if (!append) init_alphabet();


        fseek(fin,0,0);                 /* start at the beginning */
	line_no=0;

	while (fgets(line,MAXLINE+1,fin) != NULL) {
			line_no++;
		if(*line == '>') {		/* Look for start of next seq */
			fgets(line,MAXLINE+1,fin);
			line_no++;
			while (fgets(line,MAXLINE+1,fin) != NULL) { /* Look for end of seq */
			line_no++;
				if(update_alphabet(line,0)<=0) return (sint)-1;
				if(line[i]=='*') break;
			}
		}
	}

        return (sint) nseqs;
}


static sint count_msf_seqs(Boolean append)
{
/* count the number of sequences in a PILEUP alignment file */

	char line[MAXLINE+1];
	sint  nseqs;
	sint i,j,k,seqno;

	while (fgets(line,MAXLINE+1,fin) != NULL) {
			line_no++;
		if(linetype(line,"//")) break;
	}

	while (fgets(line,MAXLINE+1,fin) != NULL) {
			line_no++;
		if(!blankline(line)) break;		/* Look for next non- */
	}						/* blank line */
	nseqs = 1;

/* setup the alphabet now */
	if (!append) init_alphabet();



	while (fgets(line,MAXLINE+1,fin) != NULL) {
			line_no++;
		if(blankline(line)) break;
		nseqs++;
	}

	for(seqno=1;seqno<=nseqs;seqno++) {
	fseek(fin,0,0); 		/* start at the beginning */
	line_no=0;
	while (fgets(line,MAXLINE+1,fin) != NULL) {
			line_no++;
		if(linetype(line,"//")) break;
	}

        while (fgets(line,MAXLINE+1,fin) != NULL) {
			line_no++;
                if(!blankline(line)) {

                        for(i=1;i<seqno;i++) {
				fgets(line,MAXLINE+1,fin);
				line_no++;
			}
                        for(j=0;j<=strlen(line);j++) if(line[j] != ' ') break;
                        for(k=j;k<=strlen(line);k++) if(line[k] == ' ') break;
			if(update_alphabet(line,k)<=0) return (sint)-1;

                        for(i=0;;i++) {
                                if(fgets(line,MAXLINE+1,fin)==NULL) return nseqs;
                                if(blankline(line)) break;
                        }
                }
        }
	}

	return nseqs;
}

static sint count_rsf_seqs(Boolean append)
{
/* count the number of sequences in a GCG RSF alignment file */

	char line[MAXLINE+1];
	sint  nseqs,i,j;

	nseqs = 0;
/* skip the comments */
	while (fgets(line,MAXLINE+1,fin) != NULL) {
				line_no++;
 		if(line[strlen(line)-2]=='.' &&
                                 line[strlen(line)-3]=='.')
			break;
	}

	while (fgets(line,MAXLINE+1,fin) != NULL) {
				line_no++;
                if( *line == '{' )
                      nseqs++;
	}

	if (nseqs==0) return (sint)nseqs;

/* setup the alphabet now */
	if (!append) init_alphabet();
	fseek(fin,0,0); 		/* start at the beginning */
	line_no=0;
	

        while(fgets(line,MAXLINE+1,fin)!= NULL) {
				line_no++;
              while( !keyword(line,"sequence") )
                        if(fgets(line,MAXLINE+1,fin)==NULL) return nseqs;

              while(fgets(line,MAXLINE+1,fin)!= NULL) {
				line_no++;
			if(line[0] == '}' || line[0] == EOS)  break;
			if(update_alphabet(line,0)<=0) return (sint)-1;
	}
	}
	return (sint) nseqs;
}

static sint count_clustal_seqs(Boolean append)
/* count the number of sequences in a clustal alignment file */
{
	char line[MAXLINE+1];
	sint  nseqs;
	static char tseq[MAXLINE+1];
	static char sname[MAXLINE+1];
	sint i,j,seqno;


	while (fgets(line,MAXLINE+1,fin) != NULL) {
				line_no++;
		if(!cl_blankline(line)) break;		/* Look for next non- */
	}						/* blank line */
	nseqs = 1;

	while (fgets(line,MAXLINE+1,fin) != NULL) {
				line_no++;
		if(cl_blankline(line)) break;
		nseqs++;
	}

	if (nseqs==0) return (sint)nseqs;


/* setup the alphabet now */
	if (!append) init_alphabet();
	
	for(seqno=1;seqno<=nseqs;seqno++) {
	fseek(fin,0,0); 		/* start at the beginning */
	line_no=0;

	fgets(line,MAXLINE+1,fin);	/* read the title line...ignore it */
				line_no++;

	
	while (fgets(line,MAXLINE+1,fin) != NULL) {
				line_no++;
		if(!cl_blankline(line)) {

			for(i=1;i<seqno;i++) {
				fgets(line,MAXLINE+1,fin);
				line_no++;
			}
			for(j=0;j<=strlen(line);j++) if(line[j] != ' ') break;

			sscanf(line,"%s%s",sname,tseq);

			if(update_alphabet(tseq,0)<=0) return (sint)-1;

			for(i=0;;i++) {
				if(fgets(line,MAXLINE+1,fin)==NULL) return nseqs;
				if(cl_blankline(line)) break;
			}
		}
	}
	}

	return (sint) nseqs;
}

static void p_encode(DCHAR *seq, ECHAR *naseq, sint l)
{				/* code seq as ints .. use gap_pos2 for gap */
	register sint i;
	
	for(i=1;i<=l;i++)
		if(seq[i].c[0] == '-')
			naseq[i] = gap_pos2;
		else
			naseq[i] = res_index(alphabet,seq[i]);
	naseq[i] = -3;
}


static ECHAR res_index(DCHAR *t,DCHAR c)
{
	register ECHAR i;
	
	for(i=0;t[i].c[0] && !cmpchar(t[i],c);i++)
		;
	if(t[i].c[0]) return(i);
	else return (ECHAR)-1;
}
